/* $Id: generic.c,v 1.12 1999/02/05 00:58:25 marcus Exp $
******************************************************************************

   Generic KGI monitor driver

   Copyright (C) 1995-1997 Steffen Seeger	[seeger@ggi-project.org]
   Copyright (C) 1998      Marcus Sundberg	[marcus@ggi-project.org]

   Permission is hereby granted, free of charge, to any person obtaining a
   copy of this software and associated documentation files (the "Software"),
   to deal in the Software without restriction, including without limitation
   the rights to use, copy, modify, merge, publish, distribute, sublicense,
   and/or sell copies of the Software, and to permit persons to whom the
   Software is furnished to do so, subject to the following conditions:

   The above copyright notice and this permission notice shall be included in
   all copies or substantial portions of the Software.

   THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
   IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
   FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.  IN NO EVENT SHALL
   THE AUTHOR(S) BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER
   IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
   CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.

******************************************************************************
*/

#define	MAINTAINER	"Marcus Sundberg [marcus@ggi-project.org]"
#define DRIVER_NAME	"KGI monitor driver"
#define DRIVER_REV	"$Revision: 1.12 $"

#include <kgi/module.h>
#include "generic.h"
#include "default.h"

/* Local variables */
static struct kgi_monitor monitor               = DEF_MONITOR;
static struct kgi_preset  presets[MAX_PRESETS]  = DEF_TIMING;

/* Changeable parameters */

static int param_vmin;
static int param_vmax;
static int param_hmin;
static int param_hmax;
static int param_cmax;
static int param_xres;
static int param_yres;

static kgi_subparam param_table[] = {
	{KGIMON_VMIN, "vmin", &param_vmin, sizeof(int),
	 KGI_SUBPARAM_INT, 0644},
	{KGIMON_VMAX, "vmax", &param_vmax, sizeof(int),
	 KGI_SUBPARAM_INT, 0644},
    	{KGIMON_HMIN, "hmin", &param_hmin, sizeof(int),
	 KGI_SUBPARAM_INT, 0644},
	{KGIMON_HMAX, "hmax", &param_hmax, sizeof(int),
	 KGI_SUBPARAM_INT, 0644},
	{KGIMON_CMAX, "bandwidth", &param_cmax, sizeof(int),
	 KGI_SUBPARAM_INT, 0644},
	{KGIMON_XRES, "xres", &param_xres, sizeof(int),
	 KGI_SUBPARAM_INT, 0644},
	{KGIMON_YRES, "yres", &param_yres, sizeof(int),
	 KGI_SUBPARAM_INT, 0644},
	{0}
};

static inline void update_params(void)
{
	param_vmin = monitor.vfreq.min;
	param_vmax = monitor.vfreq.max;
	param_hmin = monitor.hfreq.min;
	param_hmax = monitor.hfreq.max;
	param_cmax = monitor.dclk.max;
	param_xres = monitor.maxdots.x;
	param_yres = monitor.maxdots.y;
}
static inline void update_monitor(void)
{
	monitor.vfreq.min = param_vmin;
	monitor.vfreq.max = param_vmax;
	monitor.hfreq.min = param_hmin;
	monitor.hfreq.max = param_hmax;
	monitor.dclk.max  = param_cmax;
	monitor.maxdots.x = param_xres;
	monitor.maxdots.y = param_yres;
}


/* Internal functions */

static int
card_specific_hacks(struct kgi_display *dpy, struct kgi_mode *mode)
{
	/* !!! Hack for the Mystique */
	if (!strcmp(dpy->card->model,"Mystique") && 
	    mode->request.graphtype == KGIGT_8BIT) {
		mode->x.total     = (mode->x.total     + 15) & ~15;
		mode->x.syncstart = (mode->x.syncstart + 15) & ~15;
		mode->x.syncend   = (mode->x.syncend   + 15) & ~15;
		return 1;
	}
	return 0;
}

static int
do_common_check(struct kgi_display *dpy, struct kgi_mode *mode)
{
	if (monitor.flags & FL_MON_PROPSIZE) {
		mode->request.size.x = mode->x.width * monitor.size.x
			/ monitor.maxdots.x;
		mode->request.size.y = mode->y.width * monitor.size.y
			/ monitor.maxdots.y;
	} else {
		mode->request.size = monitor.size;
	}

	DEBUG2("propose dclk = %i Hz, hfreq = %i Hz, vfreq = %i Hz",
	       mode->dclk, mode->dclk / mode->x.total,
	       mode->dclk / (mode->x.total * mode->y.total));

	return 0;
}


/* Multisync */

static int
multisync_check_mode(struct kgi_display *dpy, struct kgi_mode *mode,
			  enum kgi_tm_cmd *cmd)
{ 
	int hfreq, vfreq, result;
	int width = mode->x.width;

	switch (*cmd) {
	case CMD_PROPOSE:
		mode->dclk = monitor.dclk.max;

		for (hfreq = monitor.hfreq.max; hfreq > monitor.hfreq.min;
		     mode->dclk -= 1000)	{
			/*mode->x.syncstart = width * (width + 41360) / 40000;
			mode->x.syncend = width * (width * 7 / 10 + 8952) / 8000;
			mode->x.total = width * (width + 9360) / 8000;*/
			mode->x.syncstart = width * (1000 + 50) / 1000;
			mode->x.syncend = width * (1000 + 180) / 1000;
			mode->x.total = width * (1000 + 350) / 1000;
			
			card_specific_hacks(dpy, mode);

			mode->x.blankstart = mode->x.syncstart;
			mode->x.blankend = mode->x.syncend;

			hfreq = mode->dclk / mode->x.total;
			if (monitor.hfreq.max < hfreq)
				continue;

			mode->y.syncstart = mode->y.width + 3;
			mode->y.syncend   = mode->y.width + 9;
			mode->y.total     = mode->y.width + hfreq / 1000 + 1; /* 1ms flyback */
			mode->y.blankstart= mode->y.syncstart; 
			mode->y.blankend  = mode->y.syncend;

			mode->x.polarity  = mode->y.polarity = 1; /* Doesn't matter */

			vfreq = mode->dclk / (mode->x.total * mode->y.total);

			if ((monitor.vfreq.min <= vfreq) && 
			    (vfreq <= monitor.vfreq.max)) {
				
				do_common_check(dpy, mode);

				*cmd = CMD_LOWER;
				return EOK;
			}
		}
		
		NOTICE1("Invalid mode request.");
		return -E(MONITOR,UNKNOWN);
		break;
		
	case CMD_LOWER:
		if (mode->dclk < monitor.dclk.min) {
			return -E(MONITOR,UNKNOWN);
		}

		*cmd = CMD_CHECK;
		return EOK;

	case CMD_RAISE:
		if (mode->dclk > monitor.dclk.max) {
			return -E(MONITOR,UNKNOWN);
		}
		
		*cmd = CMD_CHECK;
		return EOK;
		
	case CMD_CHECK: 
		hfreq = mode->dclk / mode->x.total;
		vfreq = mode->dclk / (mode->x.total * mode->y.total);
		result =
			(monitor.hfreq.min <= hfreq) &&
			(hfreq <= monitor.hfreq.max) && 
			(monitor.vfreq.min <= vfreq) && 
			(vfreq <= monitor.vfreq.max) && 
			(monitor.dclk.min <= mode->dclk) && 
			(mode->dclk <= monitor.dclk.max);
		if (result) {
			DEBUG2("Accepted %i Hz refresh (%i Hz, DCLK=%i Hz)",
			       vfreq, hfreq, mode->dclk);
		} else {
			DEBUG2("Refresh check failed. (%i Hz, %i Hz, %i Hz)",
			       vfreq, hfreq, mode->dclk);
		}
		
		*cmd = CMD_READY;
		return result ? EOK : -E(MONITOR,UNKNOWN);
		break;

	default:
		INTERNAL_ERROR;
		return -E(MONITOR,UNKNOWN);
	}
}


/* Timelist */

static int
find_preset(int xwidth, int ywidth)
{
	int i;

	for (i = 0; i < MAX_PRESETS; i++) {
		if (presets[i].dclk == 0) {
			return NO_PRESET;
		}
		if (presets[i].x.width == xwidth 
		    && presets[i].y.width == ywidth) {
			return i;
		}
	}
	return NO_PRESET;
} 

static int
set_preset(struct kgi_preset *pres)
{
	int i = 0;

	if (! pres->dclk) {
		return -EINVAL;
	}
 
	i = find_preset(pres->x.width, pres->y.width);

	if (i == NO_PRESET) {
		/*	No entry to be replaced, thus search for a free slot.
		*/
		for (i = 0; i < MAX_PRESETS; i++) {
			if (presets[i].dclk == 0) {
				break;
			}
		}
		if (i == NO_PRESET) {
			return -ENOMEM;
		}
	}

	presets[i] = *pres;
	return EOK;
}

static int
get_preset(struct kgi_preset *pres)
{
	int i = find_preset(pres->x.width, pres->y.width);

	if (i == NO_PRESET) {
		return -EINVAL;
	} else {
		memcpy(pres, &(presets[i]), sizeof(presets[i]));
		return EOK; 
	}
}

static int
clr_preset(struct kgi_preset *pres)
{
	int i = find_preset(pres->x.width, pres->y.width);

	if (i != NO_PRESET) {
		memset(&(presets[i]), 0, sizeof(presets[i]));
	}
	
	return EOK;
}

/* Exported functions */

int
kgim_monitor_check_mode(struct kgi_display *dpy, struct kgi_mode *mode,
			enum kgi_tm_cmd *cmd)
{
	int hfreq, vfreq, h, i = 0;
	struct kgi_mode tmp = *mode;
	static struct kgi_mode save;

	update_monitor();
	if ((mode->x.width > monitor.maxdots.x) || 
	    (mode->y.width > monitor.maxdots.y)) {
		DEBUG2("Resolution too high (%ix%i).",
		       mode->x.width, mode->y.width);
		return -E(MONITOR, UNKNOWN);
	}

	if ((i = find_preset(mode->x.width, mode->y.width)) == NO_PRESET) {
		if ((monitor.synctype & SYNC_MULTISYNC)) {
			/* Try the multisync driver */
			return multisync_check_mode(dpy, mode, cmd);
		}
		return -E(MONITOR, UNKNOWN);
	}

	switch (*cmd) {

	case CMD_PROPOSE:		/* Enter timing found */
		mode->dclk = presets[i].dclk;
		mode->x    = presets[i].x;
		mode->y    = presets[i].y;

		card_specific_hacks(dpy, mode);

		do_common_check(dpy, mode);

		/*	NOTE: This depends on the kgim_check_mode()
		**	implementation!
		*/
		save = *mode;
		*cmd = CMD_LOWER;
		return EOK;

	case CMD_LOWER:
		tmp = save;
		save = *mode;
		*mode = tmp;
		*cmd  = CMD_RAISE;
		return EOK;

	case CMD_RAISE:
		/*	Take the timing whose DCLK is closer to the preset
		*/
		if (abs(save.dclk - presets[i].dclk)
			< abs(mode->dclk - presets[i].dclk)) {

			*mode = save;
		}

		memset(&save, 0, sizeof(save));
		*cmd = CMD_CHECK;
		return EOK;

	case CMD_CHECK:
		if ((mode->dclk < monitor.dclk.min) || 
		    (monitor.dclk.max < mode->dclk)) {

			DEBUG2("DCLK of %i Hz is out of bounds.", mode->dclk);
			return -E(MONITOR, UNKNOWN);
		}

		if (!(mode->x.total && mode->y.total)) {

			return -E(MONITOR, UNKNOWN);
		}

		hfreq = mode->dclk / mode->x.total;
		vfreq = hfreq / mode->y.total;
		if ((100*abs(mode->dclk-presets[i].dclk) > 2*presets[i].dclk)
		    || (hfreq < monitor.hfreq.min)
		    || (monitor.hfreq.max < hfreq)
		    || (vfreq < monitor.vfreq.min)
		    || (monitor.vfreq.max < vfreq)) {
			DEBUG2("timing limits violated, "
			       "hfreq = %i, vfreq = %i, dclk = %i",
			       hfreq, vfreq, mode->dclk);
			return -E(MONITOR, UNKNOWN);
		}
	
		h = (mode->request.graphtype == KGIGT_TEXT16) 
			? (mode->request.dpp.x)
			: (mode->request.dpp.x ? mode->request.dpp.x : 1);

#define	ERR(X) ((mode->y.X != presets[i].y.X)  \
		|| (abs(mode->x.X - presets[i].x.X) > h))
		if (ERR(width)   || ERR(blankstart) || ERR(syncstart)
		    || ERR(syncend) || ERR(blankend)   || ERR(total)) {
			NOTICE2("Error in timings.");
			return -E(MONITOR, UNKNOWN);
		}
#undef ERR
		*cmd = CMD_READY;
		return EOK;

	default:
		INTERNAL_ERROR;
		return -E(MONITOR, UNKNOWN);
	}
}


int
kgim_monitor_init(struct kgi_card *card, struct kgi_display *dpy)
{
	NOTICE(DRIVER_NAME " rev " DRIVER_REV);
	DEBUG1("%s %s (%dx%d pxl, %d x %d mm, %d-%d kHz, %d-%d Hz)",
	       monitor.manufact, monitor.model,
	       monitor.maxdots.x, monitor.maxdots.y,
	       monitor.size.x, monitor.size.y,
	       monitor.hfreq.min / 1000, monitor.hfreq.max / 1000, 
	       monitor.vfreq.min, monitor.vfreq.max);
	
	update_params();
	card->monitor = &monitor;

	return EOK;
}

void
kgim_monitor_done(struct kgi_display *dpy)
{
	DEBUG1("%s %s Monitor driver removed.", 
	       monitor.manufact, monitor.model);
}

void
kgim_monitor_set_mode(struct kgi_display *dpy, struct kgi_mode *mode)
{
}

int
kgim_monitor_command(struct kgi_graphic *state, kgi_uint cmd, kgi_uint arg)
{
	int i, result;
	
	switch(cmd) {
	case MONITOR_GETINFO:
		state->io_res = &monitor;
		return EOK;

	case MONITOR_SETINFO:
		if (suser()) {
			memcpy(&monitor, state->io_buf, sizeof(monitor));
			update_params();
			return EOK;
		} else {
			return -EPERM;
		}

	case MONITOR_GETCURTIMING: {
		struct kgi_preset *pres = (struct kgi_preset *) state->io_buf;

		pres->dclk = state->dev.dpy->mode.dclk;
		pres->x = state->dev.dpy->mode.x;
		pres->y = state->dev.dpy->mode.y;
		state->io_res = state->io_buf;

		return EOK;
	}
	
	case MONITOR_GETNUMPRESET:
		return MAX_PRESETS;
		
	case MONITOR_GETFREEPRESET:
		for (i = result = 0; i < MAX_PRESETS; i++)
			result += presets[i].dclk ? 0 : 1;
		return result;
		
	case MONITOR_SETPRESETMODE:
		if (!suser()) {
			return -EPERM;
		}
		return set_preset((struct kgi_preset *) state->io_buf);
		
	case MONITOR_GETPRESETMODE:
		return get_preset((struct kgi_preset *) state->io_res =
				  state->io_buf);
		
	case MONITOR_CLEARPRESETENT:
		if (!suser()) {
			return -EPERM;
		}
		return clr_preset((struct kgi_preset *) state->io_buf);
		
	case MONITOR_CLEARPRESETS:
		if (!suser()) {
			return -EPERM;
		}
		memset(&presets, 0, sizeof(presets));
		return EOK;
	}
	
	return -E(MONITOR, UNKNOWN);
}
	
struct kgi_subparam *kgim_monitor_get_params(void)
{
	return param_table;
}
